/*
     Sync a LED pattern over ESP NOW
     UNSECURED
     Link/description: https://werner.rothschopf.net/microcontroller/202403_esp_now_data_transmission_en.htm

     ESP32 Core 2.0.14

     2024-03-19 © noiasca

     The software is provided by the creator under fair use principle - see license.txt
*/

#include <WiFi.h>              // from ESP32 Core
#include <esp_now.h>           // from ESP32 Core

/* ***************************
    Configuration/Konfiguraton
 * ************************* */

//#define CLIENT 1             // comment/deactivate this line if you want to compile the SERVER

//                        0   1   2   3   4   5   6   7   8   9
const uint8_t ledPin[] = {2,  4, 15, 33, 32, 26, 25, 17, 16, 22};              // currently maximum are 16 pins - first pin is index 0 -> the LSB (the right most) bit in the pattern

// just some time constants used in the pattern
constexpr uint16_t tA = 40;  
constexpr uint16_t tB = 60; 
constexpr uint16_t tC = 80;
constexpr uint16_t tD = 200;
constexpr uint16_t tE = 500;
constexpr uint16_t tF = 800;
constexpr uint16_t tG = 1000;

// each line of the pattern consists of these components
struct Pattern {
  const uint16_t interval;             // the milliseconds how long the definied LEDs should be shown
  const uint16_t led;                  // indicate which of the 16 LEDs to be activated
};

// the LED pattern to be displayed, each bit defines one LED state, "paint" your pattern/sequence
const Pattern pattern[] = {
  //     5432109876543210
  {tD, 0b0000000000000000},           // all off
  {tG, 0b0000001111111111},           // all on
  {tD, 0b0000000000000000},
  {tD, 0b0000010101010101},           // several LEDs  are on
  {tD, 0b0000001010101010},
  {tD, 0b0000010101010101},
  {tD, 0b0000001010101010},
  {tD, 0b0000000000000000},
  {tC, 0b0000000000000001},
  {tC, 0b0000000000000011},
  {tC, 0b0000000000000101},
  {tC, 0b0000000000001001},
  {tC, 0b0000000000010001},
  {tC, 0b0000000000100001},
  {tC, 0b0000000001000001},
  {tC, 0b0000000010000001},
  {tC, 0b0000000100000001},
  {tD, 0b0000001000000001},
  {tC, 0b0000000100000001},
  {tC, 0b0000000010000001},
  {tC, 0b0000000001000001},
  {tC, 0b0000000000100001},
  {tC, 0b0000000000010001},
  {tC, 0b0000000000001001},
  {tC, 0b0000000000000101},
  {tC, 0b0000000000000011},
  {tD, 0b0000000000000001},
};

/* ***************************
    Globals and Functions
 * ************************* */

uint32_t lastMillis = 0;                                             // last update of the LED pattern
uint16_t actualPattern = 0;                                          // actual active pattern
uint16_t actualInterval = 0;                                         // actual interval
const size_t totalNoPattern = sizeof(pattern) / sizeof(pattern[0]);  // how many patterns are defined
const size_t totalNoPins = sizeof(ledPin) / sizeof(ledPin[0]);       // how many pins are defined

#ifdef CLIENT
void clientTransmit();                                               // prototype - this function is in another .h
#endif 

// init local hardware
void localHardwareSetup() {
  for (uint8_t i = 0; i < totalNoPins; i++) {
    pinMode(ledPin[i], OUTPUT);
  }
}

// set pattern to a specific index position and switch on/off the LEDs
void set(uint16_t index) {
  for (uint8_t i = 0; i < totalNoPins; i++) {
    if (pattern[actualPattern].led & (1 << i)) {
      digitalWrite(ledPin[i], HIGH);
    }
    else {
      digitalWrite(ledPin[i], LOW);
    }
  }
}

// run animation
void runPattern() {
  if (millis() - lastMillis >= actualInterval) {
    actualPattern++;
    if (actualPattern >= totalNoPattern) actualPattern = 0;
#ifdef CLIENT
    // on client side (sender): trigger partner. Here: only each 10th iteration
    if (actualPattern % 10 == 0) clientTransmit();
#endif
    set(actualPattern);
    actualInterval = pattern[actualPattern].interval;
    lastMillis = millis();
  }
}

/* ****************************
  ESP NOW communication
 * *************************** */

// the structure defines what data should be send/received
struct Message {
  uint16_t id = 0;           // you might send an identifier for the board or an indicator for the mode of operation
  uint16_t index = 0;        // current index of pattern (actual pattern)
};
Message messageData;

#ifdef CLIENT
#include "esp_now_client.h"
#else
#include "esp_now_server.h"
#endif

/* ****************************
  Default Arduino setup and loop
 * *************************** */
void setup() {
  Serial.begin(115200);
  delay(100);
  localHardwareSetup();
  Serial.println("\nPattern Syncronization");
  Serial.println(__DATE__);
  Serial.print("MAC: "); Serial.println(WiFi.macAddress());
  WiFi.mode(WIFI_STA);
#ifdef CLIENT
  Serial.println("Setup as SENDER");
  clientSetup();
#else
  Serial.println("Setup as RECEIVER");
  serverSetup();
#endif
}

void loop() {
  runPattern();
}
//
